@@ -24,7 +24,7 @@ from .utils import (
     _split_fvalues,
     RE_FEATURE,
     RE_FVALUE,
-    startswithset,
+    match_initial,
 )
 
 # TODO: extend documentation after blog post
@@ -91,7 +91,7 @@ class PhonoModel:
             for row in csv.DictReader(csvfile):
                 # Extract and clean strings as much as we can
                 feature = row["FEATURE"].strip()
-                fvalue = row["VALUE"].strip()  # TODO: change files to FVALUE
+                fvalue = row["FVALUE"].strip()
                 rank = int(row["RANK"].strip())
 
                 # Run checks
@@ -108,15 +108,13 @@ class PhonoModel:
                 self.features[feature].add(fvalue)
 
                 # Store values structs, which includes parsing the diacritics and
-                # the constraint string
+                # the constraint string (`constr` will be an empty list if
+                # `constraint_str` is empty)
                 prefix = replace_codepoints(row["PREFIX"])
                 suffix = replace_codepoints(row["SUFFIX"])
 
                 constraint_str = row.get("CONSTRAINTS")
-                if constraint_str:
-                    constr = parse_constraints(constraint_str)
-                else:
-                    constr = []
+                constr = parse_constraints(constraint_str)
 
                 self.fvalues[fvalue] = {
                     "feature": feature,
@@ -153,13 +151,16 @@ class PhonoModel:
         """
 
         # Load the the descriptions as an internal dictionary; we also
-        # normalize the grapheme by default
+        # normalize the grapheme by default; note that here we only sort for
+        # comparison, alphabetically, with the actual rank sorting only performed
+        # at the end if all checks pass
         with open(model_path / "sounds.csv") as csvfile:
             _graphemes = {}
             for row in csv.DictReader(csvfile):
                 grapheme = normalize(row["GRAPHEME"])
-                desc = self.sort_fvalues(row["DESCRIPTION"])
-                _graphemes[grapheme] = desc
+                _graphemes[grapheme] = self.sort_fvalues(
+                    row["DESCRIPTION"], no_rank=True
+                )
 
                 if row["CLASS"] == "True":
                     self._x["classes"].append(grapheme)
@@ -175,7 +176,7 @@ class PhonoModel:
                     ]
                 )
                 raise ValueError(
-                    f"At least one description {desc} is used for more than one sound ({at_fault})"
+                    f"`{desc}` is used for more than one sound ({at_fault})"
                 )
 
         # Check for bad fvalue names
@@ -193,9 +194,7 @@ class PhonoModel:
             # Check the grapheme constraints; we can adopt the walrus operator later
             failed = self.fail_constraints(fvalues)
             if failed:
-                raise ValueError(
-                    f"Grapheme `{grapheme}` fails constraint check on {failed}"
-                )
+                raise ValueError(f"/{grapheme}/ fails constraint check on {failed}")
 
             # Update the internal catalog
             fvalues = self.sort_fvalues(fvalues)
@@ -244,8 +243,7 @@ class PhonoModel:
                     modifier.append(val)
                 elif val != best_features[feat]:
                     modifier.append(val)
-            # TODO: use default method
-            modifier = sorted(modifier, key=lambda v: self.fvalues[v]["rank"])
+            modifier = self.sort_fvalues(modifier)
 
             # Add all modifiers as diacritics whenever possible; those without a
             # diacritic are collected in an `expression` list and will be given
@@ -295,9 +293,6 @@ class PhonoModel:
             feature).
         """
 
-        # TODO: move to tuple
-        fvalues = set(fvalues)
-
         # If the feature value is already set, not need to do the whole operation,
         # including clearing the cache, so just return to confirm
         if new_fvalue in fvalues:
@@ -309,9 +304,8 @@ class PhonoModel:
         prev_fvalue = None
         if new_fvalue[0] == "-":
             if new_fvalue[1:] in fvalues:
-                # TODO make sure to return prev_fvalue only if it was there
-                fvalues.discard(new_fvalue[1:])
                 prev_fvalue = new_fvalue[1:]
+                fvalues = [fv for fv in fvalues if fv != prev_fvalue]
         else:
             # Get the feature related to the value, cache its previous value (if any),
             # and remove it; we set `idx` to `None` in the beginning to avoid
@@ -323,16 +317,15 @@ class PhonoModel:
                     break
 
             # Remove the previous value (if there is one) and add the new value
-            fvalues.discard(prev_fvalue)
-            fvalues.add(new_fvalue)
+            fvalues = [new_fvalue] + [fv for fv in fvalues if fv != prev_fvalue]
 
         # Run a check if so requested (default)
         if check and self.fail_constraints(fvalues):
             raise ValueError(f"FValue {new_fvalue} ({feature}) breaks a constraint")
 
-        return fvalues, prev_fvalue
+        # Sort the new `fvalues`, which also makes sure we return a tuple
+        return self.sort_fvalues(fvalues), prev_fvalue
 
-    # TODO: replace set with sorted tuple
     def parse_grapheme(self, grapheme):
         """
         Parse a grapheme according to the library standard.
@@ -348,9 +341,9 @@ class PhonoModel:
             A set with the feature values from the parsed grapheme.
         """
 
-        # Used model/cache graphemes if available
+        # Used model/cache graphemes if available; it is already a sorted tuple
         if grapheme in self._x["grapheme2fvalues"]:
-            return set(self._x["grapheme2fvalues"][grapheme])
+            return self._x["grapheme2fvalues"][grapheme]
 
         # Capture list of modifiers, if any; no need to go full regex
         if "[" in grapheme and grapheme[-1] == "]":
@@ -367,7 +360,7 @@ class PhonoModel:
         # the modifiers explicitly listed as value names are consumed at the end.
         base_grapheme = ""
         while grapheme:
-            grapheme, diacritic = startswithset(grapheme, self._x["diacritics"])
+            grapheme, diacritic = match_initial(grapheme, self._x["diacritics"])
             if not diacritic:
                 base_grapheme += grapheme[0]
                 grapheme = grapheme[1:]
@@ -381,20 +374,27 @@ class PhonoModel:
             check = mod == modifiers[-1]
             fvalues, _ = self.set_fvalue(fvalues, mod, check=check)
 
-        # TODO: sort?
+        # No need to sort, as it is already returned as a sorted tuple by .set_fvalue()
         return fvalues
 
-    def sort_fvalues(self, fvalues):
+    def sort_fvalues(self, fvalues, no_rank=False):
         """
         Sort a list of values according to the model.
 
-        In case of multiple values with the same rank, these are sorted alphabetically.
+        By default, the method sorts by rank first and, in case of multiple values with
+        the same rank, alphabetically later. It is possible to perform a simple
+        alphabetical sorting.
 
         Parameters
         ----------
         fvalues : list or str
             A list (or other iterable) of values or a string description of them. If
             a string is provided, the method will split them in the standard way.
+        no_rank : bool
+            Whether to perform a plain alphabetical sorting, not using the rank
+            information. This is convenient in some cases and better than a simple
+            Python `sorted()` operation as it takes care of splitting strings and
+            accepting both strings and lists (default: `False`).
 
         Return
         ------
@@ -405,6 +405,9 @@ class PhonoModel:
         if isinstance(fvalues, str):
             fvalues = _split_fvalues(fvalues)
 
+        if no_rank:
+            return tuple(sorted(fvalues))
+
         return tuple(sorted(fvalues, key=lambda v: (-self.fvalues[v]["rank"], v)))
 
     def feature_dict(self, fvalues):
@@ -464,7 +467,6 @@ class PhonoModel:
 
         return offending
 
-    # TODO: make sure it accepts lists as well strings
     def fvalues2graphemes(self, fvalues_str, classes=False):
         """
         Collect the set of graphemes in the model that satisfy a list of fvalues.
@@ -521,7 +523,6 @@ class PhonoModel:
 
         return pass_test
 
-    # TODO: parse general graphemes, not only in model
     def minimal_matrix(self, sounds, vector=False):
         """
         Compute the minimal feature matrix for a set of sounds or graphemes.
@@ -545,7 +546,7 @@ class PhonoModel:
         # If source is a collection of strings, assume they are graphemes from
         # the model; otherwise, just take them as lists of values
         sounds = [
-            item if not isinstance(item, str) else self._x["grapheme2fvalues"][item]
+            item if not isinstance(item, str) else self.parse_grapheme(item)
             for item in sounds
         ]
 
@@ -580,7 +581,6 @@ class PhonoModel:
 
     # While this method is to a good extend similar to `.minimal_matrix`, it is not
     # reusing its codebase as it would add an unnecessary level of abstraction.
-    # TODO: parse general graphemes, not only in model
     def class_features(self, sounds):
         """
         Compute the class features for a set of graphemes or sounds.
@@ -601,7 +601,7 @@ class PhonoModel:
         # If source is a collection of strings, assume they are graphemes from
         # the model; otherwise, just take them as lists of values
         sounds = [
-            item if not isinstance(item, str) else self._x["grapheme2fvalues"][item]
+            item if not isinstance(item, str) else self.parse_grapheme(item)
             for item in sounds
         ]
 
@@ -622,9 +622,7 @@ class PhonoModel:
 
         return features
 
-    # TODO: parse general graphemes, not only in model
-    # TODO: rename binary/categorical
-    def fvalue_vector(self, source, binary=True):
+    def fvalue_vector(self, source, categorical=False):
         """
         Build a vector representation of a sound from its fvalues.
 
@@ -636,8 +634,9 @@ class PhonoModel:
         source : str or list
             Either a string with a grapheme or a feature value tuple for the sound
             to serve as basis for the vector.
-        binary : bool
-            Whether to build a binary vector (default) or a categorical one.
+        categorical : bool
+            Whether to build a categorical vector instead of a binary one
+            (default: `False`).
 
         Return
         ------
@@ -654,12 +653,12 @@ class PhonoModel:
 
         # Get the fvalues from the grapheme if `source` is a string
         if isinstance(source, str):
-            source_fvalues = self._x["grapheme2fvalues"][source]
+            source_fvalues = self.parse_grapheme(source)
         else:
             source_fvalues = source
 
         # Collect vector data in categorical or binary form
-        if not binary:
+        if categorical:
             # First get all features that are set, and later add those that
             # are not set as `None` (it is up to the user to filter them out, if
             # not wanted)
@@ -692,11 +691,7 @@ class PhonoModel:
 
         return features, vector
 
-    # TODO: parse general graphemes, not only in model
-    # TODO: allow used defined group of sounds, not only the model
-    # TODO: rename to coarsing
-    # TODO: optionally allow to use `distance`
-    def closest_grapheme(self, fvalues, classes=True):
+    def closest_grapheme(self, source, classes=True):
         """
         Find the sound in the model that is the closest to a given value tuple.
 
@@ -704,9 +699,9 @@ class PhonoModel:
 
         Parameters
         ----------
-        fvalues : tuple
+        source : tuple or str
             A feature value tuple, usually coming from the `.values` attributed of
-            a sound.
+            a sound,  or a string with a grapheme to be parsed.
         classes : bool
             Whether to allow a grapheme marked as a class to be returned; note that,
             if a grapheme class is passed but `classes` is set to `False`, a different
@@ -720,8 +715,14 @@ class PhonoModel:
             A tuple with the values for the closest match.
         """
 
-        # Check if the grapheme happens to be already in our list, also making sure
-        # the tuple is sorted
+        # Get the features if a grapheme was passed, or, in case of a feature tuple,
+        # check if the corresponding grapheme happens to be already in our
+        # internal list
+        if isinstance(source, str):
+            fvalues = self.parse_grapheme(source)
+        else:
+            fvalues = source
+
         fvalues = self.sort_fvalues(fvalues)
         if fvalues in self._x["fvalues2grapheme"]:
             grapheme = self._x["fvalues2grapheme"][fvalues]
@@ -754,7 +755,6 @@ class PhonoModel:
 
         return grapheme, best_fvalues
 
-    # TODO: parse general graphemes, not only in model
     def distance(self, sound_a, sound_b):
         """
         Return a quantitative distance based on a seed matrix.
@@ -810,32 +810,27 @@ class PhonoModel:
             if filename.is_file():
                 self._regressor = joblib.load(filename.as_posix())
                 return
-
-            raise ValueError(f"Unable to read regressor from {filename}")
+            raise RuntimeError(f"Unable to read regressor from {filename}")
 
         # Read raw distance data and cache vectors, also allowing to
         # skip over unmapped graphemes
-        # TODO: check if all can be processed, with the new .fvalue_vector
+        # TODO: some graphemes are failing because of unknown diacritics
         raw_matrix = read_distance_matrix(matrix_file)
         vector = {}
         for grapheme in raw_matrix:
             try:
                 _, vector[grapheme] = self.fvalue_vector(grapheme)
             except KeyError:
-                print("Skipping over unmapped [%s] grapheme..." % grapheme)
+                print("Skipping over grapheme [%s]..." % grapheme)
 
         # Collect (X,y) vectors
-        # TODO: deal with the +1 -- just rebuild table
         X, y = [], []  # pylint: disable=invalid-name
         for grapheme_a in raw_matrix:
             if grapheme_a in vector:
                 for grapheme_b, dist in raw_matrix[grapheme_a].items():
                     if grapheme_b in vector:
                         X.append(vector[grapheme_a] + vector[grapheme_b])
-                        if dist == 0.0:
-                            y.append(dist)
-                        else:
-                            y.append(dist + 1.0)
+                        y.append(dist)
 
         # Train regressor; setting the random value for reproducibility
         np.random.seed(42)
